@@ -5,7 +5,7 @@
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/dired-preview
-;; Version: 0.2.0
+;; Version: 0.3.0
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: files, convenience
 
@@ -40,9 +40,11 @@
 ;; Previews are shown subject to a small delay, per ther user option
 ;; `dired-preview-delay'.
 ;;
-;; Files matching the `dired-preview-ignored-extensions-regexp' are not
-;; previewed.  The default value of that user option includes multimedia,
-;; PDFs, and EPUBs.
+;; Files matching the `dired-preview-ignored-extensions-regexp' are
+;; not previewed, though a preview window is still displayed if the
+;; user option `dired-preview-ignored-show-ignored-placeholders' is
+;; non-nil.  This is to avoid windows jumping in and out of focus as
+;; the user moves between files.
 ;;
 ;; [ In the future, we may find ways to quickly preview any file type
 ;;   without affecting the performance of Emacs.]
@@ -73,11 +75,24 @@
           "\\|iso\\|epub\\|pdf\\)")
   "Regular expression of file type extensions to not preview.
 When the value is nil, do not ignore any file: preview
-everything."
+everything.
+
+A placeholder window will be displayed even for files that are ignored,
+in order to avoid windows jumping in and out of focus.  This behaviour
+is controlled by the `dired-preview-ignored-show-ignored-placeholders'
+user option."
   :group 'dired-preview
   :type '(choice (const :tag "Do not ignore any file (preview everything)" nil)
                  (string :tag "Ignore files matching regular expression")))
 
+(defcustom dired-preview-ignored-show-ignored-placeholders t
+  "When non-nil, show a placeholder preview buffer for ignored files.
+Ignored files are controlled by the `dired-preview-ignored-extensions-regexp'
+user option."
+  :type 'boolean
+  :package-version '(dired-preview . "0.3.0")
+  :group 'dired-preview)
+
 (defcustom  dired-preview-image-extensions-regexp "\\.\\(png\\|jpg\\|jpeg\\|tiff\\)"
   "List of file extensions representing image types."
   :group 'dired-preview
@@ -88,14 +103,28 @@ everything."
   :group 'dired-preview
   :type 'natnum)
 
-(defcustom dired-preview-display-action-alist-function
+(define-obsolete-variable-alias
+  'dired-preview-display-action-alist-function
+  'dired-preview-display-action-alist
+  "0.3.0")
+
+(defcustom dired-preview-display-action-alist
   #'dired-preview-display-action-alist-dwim
-  "Function to return the `display-buffer' action for the preview.
+  "The `display-buffer' action alist for the preview window.
 This is the same data that is passed to `display-buffer-alist'.
 Read Info node `(elisp) Displaying Buffers'.  As such, it is
-meant for experienced users.  See the reference function
-`dired-preview-display-action-alist-dwim' for the implementation
-details."
+meant for experienced users.
+
+Example of a valid value:
+
+    \\='((display-buffer-in-side-window)
+      (side . bottom)
+      (window-height . 0.2)
+      (preserve-size . (t . t)))
+
+The value may also be a function, which returns a `display-buffer'
+action alist.  See `dired-preview-display-action-alist-dwim' for the
+implementation details."
   :group 'dired-preview
   :type 'function)
 
@@ -119,10 +148,6 @@ details."
   "Return buffers that show previews."
   (seq-filter #'buffer-live-p dired-preview--buffers))
 
-(defun dired-preview--window-parameter-p (window)
-  "Return non-nil if WINDOW has `dired-preview-window' parameter."
-  (window-parameter window 'dired-preview-window))
-
 ;; TODO 2023-07-07: This can become a user option, but let's keep it
 ;; simple for now.  We need to be sure this is always doing the right
 ;; thing.
@@ -164,6 +189,21 @@ until it drops below this number.")
         dired-preview--large-files-alist)
   (setq dired-preview--large-files-alist nil))
 
+(defun dired-preview--kill-placeholder-buffers ()
+  "Kill all placeholder preview buffers."
+  (setq dired-preview--buffers
+        (seq-remove (lambda (buffer)
+                      (with-current-buffer buffer
+                        (when (and (boundp 'dired-preview--placeholder-buffer-p)
+                                   dired-preview--placeholder-buffer-p)
+                          (ignore-errors (kill-buffer buffer))
+                          t)))
+                    (dired-preview--get-buffers))))
+
+(defun dired-preview--window-parameter-p (window)
+  "Return non-nil if WINDOW has `dired-preview-window' parameter."
+  (window-parameter window 'dired-preview-window))
+
 (defun dired-preview--get-windows ()
   "Return windows that show previews."
   (seq-filter #'dired-preview--window-parameter-p (window-list)))
@@ -219,6 +259,7 @@ FILE."
         (file (expand-file-name file)))
     (cond
      ((and (not (string-empty-p ext))
+           dired-preview-ignored-extensions-regexp
            (string-match-p ext dired-preview-ignored-extensions-regexp))
       (cons 'ignore file))
      ((dired-preview--file-large-p file)
@@ -254,17 +295,113 @@ FILE."
      end-ov 'display
      (propertize "\n--PREVIEW TRUNCATED--" 'face 'shadow))))
 
+;;;###autoload
+(defmacro dired-preview-with-window (&rest body)
+  "Evaluate BODY with the Dired preview window as selected."
+  (declare (indent 0))
+  `(dolist (win (dired-preview--get-windows))
+     (with-selected-window win
+       ,@body)))
+
+(defun dired-preview-find-file ()
+  "Visit the currently previewed buffer with `find-file'.
+This means that the buffer is no longer among the previews.
+
+Also see `dired-preview-open-dwim'."
+  (interactive)
+  (let (file buffer)
+    (dired-preview-with-window
+      (setq file buffer-file-name)
+      (dired-preview--close-previews-outside-dired)
+      (setq buffer (find-file-noselect file)))
+    (pop-to-buffer buffer)))
+
+(defvar dired-preview-media-extensions-regexp
+  "\\.\\(mp3\\|m4a\\|flac\\|mp4\\|ogg\\|mpv\\|webm\\|mov\\|wav\\)"
+  "Regular expression matching media file extensions.")
+
+(declare-function w32-shell-execute "w32fns.c")
+
+;; NOTE 2024-07-29: Adapted from the `dired-do-open' found in Emacs 31.0.50.
+(defun dired-preview--open-externally (file)
+  "Run appropriate command to open FILE externally."
+  (if-let ((command (cond
+                     ((executable-find "xdg-open")
+                      "xdg-open")
+                     ((memq system-type '(gnu/linux darwin))
+                      "open")
+                     ((memq system-type '(windows-nt ms-dos))
+                      "start")
+                     ((eq system-type 'cygwin)
+                      "cygstart")
+                     ((executable-find "run-mailcap")
+                      "run-mailcap"))))
+      (cond
+       ((memq system-type '(gnu/linux))
+        (call-process command nil 0 nil file))
+       ((memq system-type '(ms-dos))
+        (shell-command (concat command " " (shell-quote-argument file))))
+       ((memq system-type '(windows-nt))
+        (w32-shell-execute command (convert-standard-filename file)))
+       ((memq system-type '(cygwin))
+        (call-process command nil nil nil file))
+       ((memq system-type '(darwin))
+        (start-process (concat command " " file) nil command file)))
+    (error "Cannot find a command to open `%s' externally" file)))
+
+(defun dired-preview-open-dwim ()
+  "Do-What-I-Mean open the currently previewed file.
+This means that the buffer is no longer among the previews.
+
+If the file name matches `dired-preview-media-extensions-regexp',
+`dired-preview-ignored-extensions-regexp', or
+`dired-preview-image-extensions-regexp', then open it externally.
+Otherwise, visit the file in an Emacs buffer.
+
+Also see `dired-preview-find-file'."
+  (interactive)
+  (let (buffer)
+    (dired-preview-with-window
+      (when-let ((file buffer-file-name))
+        (cond
+         ((or (string-match-p dired-preview-media-extensions-regexp file)
+              (string-match-p dired-preview-ignored-extensions-regexp file)
+              (string-match-p dired-preview-image-extensions-regexp file))
+          (dired-preview--open-externally file))
+         (t
+          (dired-preview--close-previews-outside-dired)
+          (setq buffer (find-file-noselect file))))))
+    (when buffer
+      (pop-to-buffer buffer))))
+
+;; NOTE 2024-07-29: "Scroll up/down" confuses me in this context
+;; because the motion is in the opposite direction.  So "page up/down"
+;; is fine, based on what the keys of the same name do.
+(defun dired-preview-page-down ()
+  "Move a page down in the preview window.
+This technically runs `scroll-up-command'."
+  (interactive)
+  (dired-preview-with-window
+    (call-interactively 'scroll-up-command)))
+
+;; Same as above for the terminology.
+(defun dired-preview-page-up ()
+  "Move a page up in the preview window.
+This technically runs `scroll-down-command'."
+  (interactive)
+  (dired-preview-with-window
+    (call-interactively 'scroll-down-command)))
+
 (declare-function hexl-mode "hexl")
 (declare-function hexl-mode-exit "hexl" (&optional arg))
 
 (defun dired-preview-hexl-toggle ()
   "Toggle preview between text and `hexl-mode'."
   (interactive)
-  (dolist (win (dired-preview--get-windows))
-    (with-selected-window win
-      (if (eq major-mode 'hexl-mode)
-          (hexl-mode-exit)
-        (hexl-mode))
+  (dired-preview-with-window
+    (if (eq major-mode 'hexl-mode)
+        (hexl-mode-exit)
+      (hexl-mode)
       (dired-preview--add-truncation-message))))
 
 (cl-defmethod dired-preview--get-buffer ((file (head large)))
@@ -297,8 +434,18 @@ The size of the leading chunk is specified by
                 (current-buffer)))))))
 
 (cl-defmethod dired-preview--get-buffer ((file (head ignore)))
-  "Get preview buffer for ignored FILE."
-  (message "No preview method for `%s'" (cdr file)))
+  "Get preview placeholder buffer for an ignored FILE."
+  (let* ((file (cdr file))
+         (buffer-name (format "%s (no preview)"
+                              (file-name-nondirectory file))))
+    (or (get-buffer buffer-name)
+        (with-current-buffer (get-buffer-create buffer-name)
+          (set-visited-file-name file t)
+          (rename-buffer buffer-name)
+          (set-buffer-modified-p nil)
+          (setq-local dired-preview--placeholder-buffer-p t)
+          (setq buffer-read-only t)
+          (current-buffer)))))
 
 (cl-defmethod dired-preview--get-buffer ((file (head directory)))
   "Get preview buffer for directory FILE type."
@@ -365,7 +512,7 @@ checked against `split-width-threshold' or
     `(:side bottom :dimension window-height :size ,(dired-preview-get-window-size :height))))
 
 (defun dired-preview-display-action-alist-dwim ()
-  "Reference function for `dired-preview-display-action-alist-function'.
+  "Reference function for `dired-preview-display-action-alist'.
 Return a `display-buffer' action alist, as described in the
 aforementioned user option."
   (let ((properties (dired-preview-display-action-side)))
@@ -376,6 +523,7 @@ aforementioned user option."
 (defvar dired-preview-trigger-commands
   '( dired-next-line
      dired-previous-line
+     dired-flag-file-deletion
      dired-mark
      dired-unmark
      dired-unmark-backward
@@ -399,7 +547,8 @@ aforementioned user option."
   (dired-preview--cancel-timer)
   (dired-preview--delete-windows)
   (dired-preview--kill-buffers)
-  (dired-preview--kill-large-buffers))
+  (dired-preview--kill-large-buffers)
+  (dired-preview--kill-placeholder-buffers))
 
 (defun dired-preview--close-previews-outside-dired ()
   "Call `dired-preview--close-previews' if the current buffer is not in Dired mode."
@@ -411,10 +560,13 @@ aforementioned user option."
 (defun dired-preview--display-buffer (buffer)
   "Call `display-buffer' for BUFFER.
 Only do it with the current major mode is Dired."
-  (display-buffer
-   buffer
-   (funcall (or dired-preview-display-action-alist-function
-                #'dired-preview-display-action-alist-dwim))))
+  (let ((action-alist (cond
+                       ((functionp dired-preview-display-action-alist)
+                        (funcall dired-preview-display-action-alist))
+                       (dired-preview-display-action-alist)
+                       (t
+                        (dired-preview-display-action-alist-dwim)))))
+    (display-buffer buffer action-alist)))
 
 (defun dired-preview-display-file (file)
   "Display preview of FILE if appropriate."
@@ -430,7 +582,8 @@ Only do it with the current major mode is Dired."
        (or (file-regular-p file) (file-directory-p file))
        (file-readable-p file)
        (not (dired-preview--file-displayed-p file))
-       (not (dired-preview--file-ignored-p file))))
+       (or dired-preview-ignored-show-ignored-placeholders
+           (not (dired-preview--file-ignored-p file)))))
 
 (defun dired-preview-start (file)
   "Preview FILE instantly when invoking Dired."
@@ -477,6 +630,10 @@ the preview with `dired-preview-delay' of idleness."
 (defvar dired-preview-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "C-c C-c") #'dired-preview-hexl-toggle)
+    (define-key map (kbd "C-c C-f") #'dired-preview-find-file)
+    (define-key map (kbd "C-c C-o") #'dired-preview-open-dwim)
+    (define-key map (kbd "C-c C-u") #'dired-preview-page-up)
+    (define-key map (kbd "C-c C-d") #'dired-preview-page-down)
     map)
   "Key map for `dired-preview-mode'.")
 
